from ..board.board import Board
from .base_agent import BaseAgent

class Agent(BaseAgent):
  def __init__(self, name: str, color: str):
    self.validate(color)
    self.name = name
    self.color = color
    self.symbol = color[0].capitalize()
    print(f'{self.name} has entered the game as team {self.color}')
  
  def validate(self, color):
    valid_colors = ['red', 'blue']
    if color not in valid_colors:
      raise Exception(f'Invalid color {color}, use one of {valid_colors}')

  def place_block(self):
    column_valid = False
    while not column_valid:
      column = int(input("Choose a column (0-6): "))
      column_valid = self.board.validate_move(column)
      if column_valid:
        return self.board.place_block(column, self.symbol)
      else:
        print('Invalid column, try again')


import pdb

class Board:
  def __init__(self, column_count=7, row_count=6):
    self.column_count = column_count
    self.row_count = row_count
    self.create_grid()

  def create_grid(self):
    gridline = []
    for _ in range(self.column_count):
        gridline.append('-')
    grid = []
    for _ in range(self.row_count):
        grid.append(list(gridline))

    print(f'grid has {len(grid)} rows and {len(grid[0])} columns')
    self.grid = grid
  
  def print_board(self):
    for i in range(self.row_count):
      print(self.grid[i])
    print([f'{i}' for i in range(self.column_count)])

  def place_block(self, column, symbol):
    print(f'placing column {column} with symbol {symbol}')
    row = False
    row = self.find_next_free_row(column)
    self.grid[row][column] = symbol
    return {
      'row': row,
      'column': column,
    }
  
  def validate_move(self, column):
    if column < 0 or column > self.column_count - 1:
      return False
    row = self.find_next_free_row(column)
    if row < 0 or row > self.row_count:
      return False
    return True
  
  def find_next_free_row(self, column):
    row = -1
    for i in reversed(range(self.row_count)):
      if self.grid[i][column] == '-':
        print(f'found free row {i}')
        row = i
        break

    return row
  
  def get_row(self, row):
    return self.grid[row]

  def get_column(self, column):
    return [row[column] for row in self.grid]

  def get_cell(self, row, column):
    print(f'getting cell at row {row} and column {column}')
    # pdb.set_trace()
    return self.grid[row][column]


from ..board import Board
from . import BoardEvaluator
from ..agent import Agent
from ..setup_test_env import setup_test_env

def test_setup():
  test_env = setup_test_env()
  return {
    'board_evaluator': BoardEvaluator(test_env['lcm'].board, test_env['red_agent'], test_env['blue_agent']),
    'red_agent': test_env['red_agent'],
    'blue_agent': test_env['blue_agent'],
    'board': test_env['lcm'].board
  }

def test_can_spot_a_horizontal_win():
  setup = test_setup()
  evaluator = setup['board_evaluator']
  red_agent = setup['red_agent']
  board = setup['board']

  board.grid[0][0] = 'R'
  result = evaluator.evaluate(red_agent, {'row': 0, 'column': 0})
  assert result['completed'] == False

  board.grid[0][1] = 'R'
  result = evaluator.evaluate(red_agent, {'row': 0, 'column': 1})
  assert result['completed'] == False

  board.grid[0][2] = 'R'
  result = evaluator.evaluate(red_agent, {'row': 0, 'column': 2})
  assert result['completed'] == False

  board.grid[0][3] = 'R'
  result = evaluator.evaluate(red_agent, {'row': 0, 'column': 3})
  assert result['completed'] == True
  assert result['winner'] == red_agent

def test_does_not_count_a_horizontal_win_if_not_connected():
  setup = test_setup()
  evaluator = setup['board_evaluator']
  red_agent = setup['red_agent']
  board = setup['board']

  board.grid[0][0] = 'R'
  result = evaluator.evaluate(red_agent, {'row': 0, 'column': 0})
  assert result['completed'] == False

  board.grid[0][1] = 'R'
  result = evaluator.evaluate(red_agent, {'row': 0, 'column': 1})
  assert result['completed'] == False

  board.grid[0][2] = 'R'
  result = evaluator.evaluate(red_agent, {'row': 0, 'column': 2})
  assert result['completed'] == False

  board.grid[0][4] = 'R'
  result = evaluator.evaluate(red_agent, {'row': 0, 'column': 3})
  assert result['completed'] == False

def test_can_spot_a_vertical_win():
  setup = test_setup()
  evaluator = setup['board_evaluator']
  blue_agent = setup['blue_agent']
  board = setup['board']

  board.grid[1][0] = 'B'
  board.grid[2][0] = 'B'
  board.grid[3][0] = 'B'
  result = evaluator.evaluate(blue_agent, {'row': 3, 'column': 0})
  assert result['completed'] == False

  board.grid[4][0] = 'B'
  result = evaluator.evaluate(blue_agent, {'row': 4, 'column': 0})
  assert result['completed'] == True
  assert result['winner'] == blue_agent

def test_can_spot_a_positive_diagonal_win():
  setup = test_setup()
  evaluator = setup['board_evaluator']
  blue_agent = setup['blue_agent']
  board = setup['board']

  board.grid[1][0] = 'B'
  board.grid[2][1] = 'B'
  board.grid[3][2] = 'B'
  result = evaluator.evaluate(blue_agent, {'row': 3, 'column': 2})
  assert result['completed'] == False

  board.grid[4][3] = 'B'
  result = evaluator.evaluate(blue_agent, {'row': 4, 'column': 3})
  assert result['completed'] == True
  assert result['winner'] == blue_agent

def test_can_spot_a_negative_diagonal_win():
  setup = test_setup()
  evaluator = setup['board_evaluator']
  blue_agent = setup['blue_agent']
  board = setup['board']

  board.grid[1][3] = 'B'
  board.grid[2][2] = 'B'
  board.grid[3][1] = 'B'
  result = evaluator.evaluate(blue_agent, {'row': 3, 'column': 1})
  assert result['completed'] == False

  board.grid[4][0] = 'B'
  result = evaluator.evaluate(blue_agent, {'row': 4, 'column': 0})
  assert result['completed'] == True
  assert result['winner'] == blue_agent


from .agent.agent import Agent
from .lifecycle_manager.lifecycle_manager import LifeCycleManager as Game

def setup_test_env():
  red_agent = Agent('RED TEAM', 'red')
  blue_agent = Agent('BLUE TEAM', 'blue')
  lcm = Game(red_agent, blue_agent)
  return { 'red_agent': red_agent, 'blue_agent': blue_agent, 'lcm': lcm}



from ..agent.agent import Agent
from ..board.board import Board
from ..board_evaluator.board_evaluator import BoardEvaluator
from ..utils import has_method
import os

class LifeCycleManager():
  game_finished = False
  winner = None
  def __init__(self, red_agent, blue_agent, row_count=6, column_count=7):
    self.board = Board(row_count=row_count, column_count=column_count)
    self.grid = self.board.grid
    self.red_agent = red_agent
    self.current_turn = self.red_agent
    self.blue_agent = blue_agent
    self.red_agent.add_to_game(self)
    self.blue_agent.add_to_game(self)

  def start_game(self):
    self.evaluator = BoardEvaluator(
      self.board, self.red_agent, self.blue_agent
    )
    self.board.print_board()
    self.main_game_loop()

  def main_game_loop(self):
    while not self.game_finished:
      coords = self.make_move()
      self.evaluate_board(coords)
      self.post_evaluation_hook()

  def evaluate_board(self, coords):
    result = self.evaluator.evaluate(self.current_turn, coords)
    if result['completed']:
      self.board.print_board()
      print('Game over!')
      print(f'{result["winner"].name} has won the game!')
      self.game_finished = True
      self.winner = result['winner']
    else:
      self.current_turn = self.blue_agent if self.current_turn == self.red_agent else self.red_agent

  def make_move(self):
    os.system('clear')
    agent = self.current_turn
    self.board.print_board()
    print(f'{agent.name}\'s turn ({agent.color})')
    return agent.place_block()

  def post_evaluation_hook(self):
    if has_method(self.red_agent, 'post_evaluation_hook'):
      self.red_agent.post_evaluation_hook()
    if has_method(self.blue_agent, 'post_evaluation_hook'):
      self.blue_agent.post_evaluation_hook()




class BoardEvaluator():
  def __init__(self, board, red_agent, blue_agent):
    self.board = board
    self.winner = None
    self.max_row = board.row_count - 1
    self.max_column = board.column_count - 1

  def evaluate(self, agent, coords):
    self.board.print_board()
    if not agent or not coords:
      raise Exception(
        f'Invalid evaluation, missing agent or coords - agent: {agent}, coords: {coords}'
      )
    self.current_agent = agent
    self.current_coords = coords
    self.current_symbol = agent.symbol
    self.check_for_horizontal_win()
    self.check_for_vertical_win()
    self.check_for_diagonal_win()
    self.is_draw()

    return {
      'completed': self.is_completed(),
      'winner': self.winner,
    }

  def check_for_horizontal_win(self):
    if self.winner or self.is_draw():
      return
    row = self.board.get_row(self.current_coords["row"])
    self.check_array_for_win(row)

  def check_array_for_win(self, array):
    current_count = 0
    for i in array:
      if current_count == 4:
        break
      if i == self.current_symbol:
        current_count += 1
      else:
        current_count = 0
    if current_count == 4:
      self.winner = self.current_agent

  def check_for_vertical_win(self):
    if self.winner or self.is_draw():
      return
    column = self.board.get_column(self.current_coords["column"])
    self.check_array_for_win(column)
  
  def check_for_diagonal_win(self):
    if self.winner or self.is_draw():
      return
    positive_diagonal = self.create_positive_diagonal_array(self.current_coords["row"], self.current_coords["column"])
    negative_diagonal = self.create_negative_diagonal_array(self.current_coords["row"], self.current_coords["column"])
    self.check_array_for_win(positive_diagonal)
    self.check_array_for_win(negative_diagonal)

  def create_positive_diagonal_array(self, row, column):
    diagonal = []
    while row > 0 and column > 0:
      row -= 1
      column -= 1
    
    while row < (self.max_row) and column < (self.max_column):
      diagonal.append(self.board.get_cell(row, column))
      row += 1
      column += 1
    return diagonal

  def create_negative_diagonal_array(self, row, column):
    diagonal = []
    min_row = row
    max_column = column
    while min_row > 0 and max_column < (self.max_column):
      min_row -= 1
      max_column += 1
    
    while min_row < (self.max_row) and max_column >= 0:
      diagonal.append(self.board.get_cell(min_row, max_column))
      min_row += 1
      max_column -= 1
    return diagonal

  def is_completed(self):
    return self.is_draw() or bool(self.winner)

  def is_draw(self):
    return False


from . import Agent
from ..board import Board
from ..setup_test_env import setup_test_env

def test_can_place_block(monkeypatch):
  test_env = setup_test_env()
  agent = test_env['red_agent']

  monkeypatch.setattr('builtins.input', lambda _: "1")
  output = agent.place_block()
  assert output == { 'row': 5, 'column': 1 }
  output = agent.place_block()
  assert output == { 'row': 4, 'column': 1 }